@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2020-2022 Fanout, Inc.
+ * Copyright (C) 2020-2023 Fanout, Inc.
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -21,6 +21,7 @@ use std::io;
 use std::io::{Read, Write};
 use std::mem::{self, MaybeUninit};
 use std::rc::Rc;
+use std::slice;
 
 pub const VECTORED_MAX: usize = 8;
 
@@ -55,6 +56,7 @@ pub fn trim_for_display(s: &str, max: usize) -> String {
     }
 }
 
+#[allow(clippy::len_without_is_empty)]
 pub trait RefRead {
     fn len(&self) -> usize;
     fn get_ref(&self) -> &[u8];
@@ -65,16 +67,16 @@ pub trait RefRead {
         &'data self,
         bufs: &'bufs mut [&'data [u8]],
     ) -> &'bufs mut [&'data [u8]] {
-        assert!(bufs.len() >= 1);
+        assert!(!bufs.is_empty());
 
         bufs[0] = self.get_ref();
 
         &mut bufs[..1]
     }
 
-    fn get_mut_vectored<'data, 'bufs>(
+    fn get_mut_vectored<'data, 'bufs, const N: usize>(
         &'data mut self,
-        bufs: &'bufs mut MaybeUninit<[&'data mut [u8]; VECTORED_MAX]>,
+        bufs: &'bufs mut MaybeUninit<[&'data mut [u8]; N]>,
     ) -> &'bufs mut [&'data mut [u8]] {
         let bufs = unsafe { bufs.assume_init_mut() };
 
@@ -108,8 +110,8 @@ impl RefRead for io::Cursor<&mut [u8]> {
     }
 }
 
-pub fn write_vectored_offset(
-    writer: &mut dyn Write,
+pub fn write_vectored_offset<W: Write>(
+    writer: &mut W,
     bufs: &[&[u8]],
     offset: usize,
 ) -> Result<usize, io::Error> {
@@ -138,12 +140,8 @@ pub fn write_vectored_offset(
     let mut arr = [io::IoSlice::new(&b""[..]); VECTORED_MAX];
     let mut arr_len = 0;
 
-    for i in start..bufs.len() {
-        let buf = if i == start {
-            &bufs[i][offset..]
-        } else {
-            bufs[i]
-        };
+    for (index, &buf) in bufs.iter().enumerate().skip(start) {
+        let buf = if index == start { &buf[offset..] } else { buf };
 
         arr[arr_len] = io::IoSlice::new(buf);
         arr_len += 1;
@@ -152,8 +150,8 @@ pub fn write_vectored_offset(
     writer.write_vectored(&arr[..arr_len])
 }
 
-pub async fn write_vectored_offset_async(
-    writer: &mut dyn AsyncWrite,
+pub async fn write_vectored_offset_async<W: AsyncWrite>(
+    writer: &mut W,
     bufs: &[&[u8]],
     offset: usize,
 ) -> Result<usize, io::Error> {
@@ -182,12 +180,8 @@ pub async fn write_vectored_offset_async(
     let mut arr = [io::IoSlice::new(&b""[..]); VECTORED_MAX];
     let mut arr_len = 0;
 
-    for i in start..bufs.len() {
-        let buf = if i == start {
-            &bufs[i][offset..]
-        } else {
-            bufs[i]
-        };
+    for (index, &buf) in bufs.iter().enumerate().skip(start) {
+        let buf = if index == start { &buf[offset..] } else { buf };
 
         arr[arr_len] = io::IoSlice::new(buf);
         arr_len += 1;
@@ -196,58 +190,192 @@ pub async fn write_vectored_offset_async(
     writer.write_vectored(&arr[..arr_len]).await
 }
 
-pub trait LimitBufs<'a> {
-    fn limit(&mut self, size: usize) -> &mut [&'a [u8]];
+struct LimitBufsRestore<T> {
+    index: usize,
+    ptr: T,
+    len: usize,
 }
 
-impl<'a> LimitBufs<'a> for [&'a [u8]] {
-    fn limit(&mut self, size: usize) -> &mut [&'a [u8]] {
-        let mut want = size;
+pub struct LimitBufsGuard<'a, 'b> {
+    bufs: &'b mut [&'a [u8]],
+    start: usize,
+    end: usize,
+    restore: Option<LimitBufsRestore<*const u8>>,
+}
 
-        for i in 0..self.len() {
-            let buf_len = self[i].len();
-
-            if buf_len >= want {
-                self[i] = &self[i][..want];
-                return &mut self[..(i + 1)];
-            }
-
-            want -= buf_len;
-        }
-
-        self
+impl<'a: 'b, 'b> LimitBufsGuard<'a, 'b> {
+    pub fn as_slice(&self) -> &[&'a [u8]] {
+        &self.bufs[self.start..self.end]
     }
 }
 
-pub trait LimitBufsMut<'a> {
-    fn limit(&mut self, size: usize) -> &mut [&'a mut [u8]];
+impl<'a: 'b, 'b> Drop for LimitBufsGuard<'a, 'b> {
+    fn drop(&mut self) {
+        if let Some(restore) = self.restore.take() {
+            // SAFETY: ptr and len were collected earlier from the original
+            // memory referred to by the slice at this index and they are
+            // still valid. the only issue with reconstructing the slice is
+            // that we currently have a different slice using the same memory
+            // at this index. however, this is safe because we also replace
+            // the slice at this index and the two slices don't coexist
+            unsafe {
+                self.bufs[restore.index] = slice::from_raw_parts(restore.ptr, restore.len);
+            }
+        }
+    }
 }
 
-impl<'a> LimitBufsMut<'a> for [&'a mut [u8]] {
-    fn limit(&mut self, size: usize) -> &mut [&'a mut [u8]] {
+pub struct LimitBufsMutGuard<'a, 'b> {
+    bufs: &'b mut [&'a mut [u8]],
+    start: usize,
+    end: usize,
+    restore: Option<LimitBufsRestore<*mut u8>>,
+}
+
+impl<'a: 'b, 'b> LimitBufsMutGuard<'a, 'b> {
+    pub fn as_slice(&mut self) -> &mut [&'a mut [u8]] {
+        &mut self.bufs[self.start..self.end]
+    }
+}
+
+impl<'a: 'b, 'b> Drop for LimitBufsMutGuard<'a, 'b> {
+    fn drop(&mut self) {
+        if let Some(restore) = self.restore.take() {
+            // SAFETY: ptr and len were collected earlier from the original
+            // memory referred to by the slice at this index and they are
+            // still valid. the only issue with reconstructing the slice is
+            // that we currently have a different slice using the same memory
+            // at this index. however, this is safe because we also replace
+            // the slice at this index and the two slices don't coexist
+            unsafe {
+                self.bufs[restore.index] = slice::from_raw_parts_mut(restore.ptr, restore.len);
+            }
+        }
+    }
+}
+
+pub trait LimitBufs<'a, 'b> {
+    fn limit(&'b mut self, size: usize) -> LimitBufsGuard<'a, 'b>;
+}
+
+impl<'a: 'b, 'b> LimitBufs<'a, 'b> for [&'a [u8]] {
+    fn limit(&'b mut self, size: usize) -> LimitBufsGuard<'a, 'b> {
+        let mut end = self.len();
+        let mut restore = None;
         let mut want = size;
 
-        for i in 0..self.len() {
-            let buf_len = self[i].len();
+        for (index, item) in self.iter_mut().enumerate() {
+            let buf: &[u8] = item;
+            let buf_len = buf.len();
 
             if buf_len >= want {
-                // SAFETY: shrinking the length of an existing slice requires
-                // borrowing through self, which causes the compiler to
-                // extend the lifetime of the smaller slice to the lifetime
-                // of self. however, the smaller slice is not dependent on
-                // anything about self except the original slice, so it is
-                // safe to forcibly reduce its lifetime to that of the
-                // original
-                self[i] =
-                    unsafe { mem::transmute::<&mut [u8], &'a mut [u8]>(&mut self[i][..want]) };
+                let len = buf.len();
+                let ptr = buf.as_ptr();
 
-                return &mut self[..(i + 1)];
+                restore = Some(LimitBufsRestore { index, ptr, len });
+
+                // SAFETY: ptr and len were obtained above and are still
+                // valid. we just need to be careful about using them again
+                // later on from the restore field
+                unsafe {
+                    *item = &slice::from_raw_parts(ptr, len)[..want];
+                }
+
+                end = index + 1;
+                break;
             }
 
             want -= buf_len;
         }
 
-        self
+        LimitBufsGuard {
+            bufs: self,
+            start: 0,
+            end,
+            restore,
+        }
+    }
+}
+
+pub trait LimitBufsMut<'a: 'b, 'b> {
+    fn skip(&'b mut self, size: usize) -> LimitBufsMutGuard<'a, 'b>;
+    fn limit(&'b mut self, size: usize) -> LimitBufsMutGuard<'a, 'b>;
+}
+
+impl<'a: 'b, 'b> LimitBufsMut<'a, 'b> for [&'a mut [u8]] {
+    fn skip(&'b mut self, size: usize) -> LimitBufsMutGuard<'a, 'b> {
+        let mut start = 0;
+        let end = self.len();
+        let mut restore = None;
+        let mut skip = size;
+
+        for (index, item) in self.iter_mut().enumerate() {
+            let buf: &mut [u8] = item;
+            let buf_len = buf.len();
+
+            if buf_len >= skip {
+                let len = buf.len();
+                let ptr = buf.as_mut_ptr();
+
+                restore = Some(LimitBufsRestore { index, ptr, len });
+
+                // SAFETY: ptr and len were obtained above and are still
+                // valid. we just need to be careful about using them again
+                // later on from the restore field
+                unsafe {
+                    *item = &mut slice::from_raw_parts_mut(ptr, len)[skip..];
+                }
+
+                start = index;
+                break;
+            }
+
+            skip -= buf_len;
+        }
+
+        LimitBufsMutGuard {
+            bufs: self,
+            start,
+            end,
+            restore,
+        }
+    }
+
+    fn limit(&'b mut self, size: usize) -> LimitBufsMutGuard<'a, 'b> {
+        let mut end = self.len();
+        let mut restore = None;
+        let mut want = size;
+
+        for (index, item) in self.iter_mut().enumerate() {
+            let buf: &mut [u8] = item;
+            let buf_len = buf.len();
+
+            if buf_len >= want {
+                let len = buf.len();
+                let ptr = buf.as_mut_ptr();
+
+                restore = Some(LimitBufsRestore { index, ptr, len });
+
+                // SAFETY: ptr and len were obtained above and are still
+                // valid. we just need to be careful about using them again
+                // later on from the restore field
+                unsafe {
+                    *item = &mut slice::from_raw_parts_mut(ptr, len)[..want];
+                }
+
+                end = index + 1;
+                break;
+            }
+
+            want -= buf_len;
+        }
+
+        LimitBufsMutGuard {
+            bufs: self,
+            start: 0,
+            end,
+            restore,
+        }
     }
 }
 
@@ -343,6 +471,7 @@ impl Write for Buffer {
 
 pub struct TmpBuffer(RefCell<Vec<u8>>);
 
+#[allow(clippy::len_without_is_empty)]
 impl TmpBuffer {
     pub fn new(size: usize) -> Self {
         Self(RefCell::new(vec![0; size]))
@@ -399,7 +528,7 @@ impl<T: AsRef<[u8]> + AsMut<[u8]>> BaseRingBuffer<T> {
         self.end = 0;
     }
 
-    pub fn write_from(&mut self, r: &mut dyn Read) -> Result<usize, io::Error> {
+    pub fn write_from<R: Read>(&mut self, r: &mut R) -> Result<usize, io::Error> {
         let size = match r.read(self.write_buf()) {
             Ok(size) => size,
             Err(e) => return Err(e),
@@ -525,7 +654,7 @@ impl<T: AsRef<[u8]> + AsMut<[u8]>> BaseRingBuffer<T> {
 
             tmp[..lsize].copy_from_slice(&buf[lsrc..(lsrc + lsize)]);
             buf.copy_within(hsrc..(hsrc + hsize), hdest);
-            buf[ldest..(ldest + lsize)].copy_from_slice(&mut tmp[..lsize]);
+            buf[ldest..(ldest + lsize)].copy_from_slice(&tmp[..lsize]);
         }
 
         self.start = 0;
@@ -609,7 +738,7 @@ impl<T: AsRef<[u8]> + AsMut<[u8]>> RefRead for BaseRingBuffer<T> {
         &'data self,
         bufs: &'bufs mut [&'data [u8]],
     ) -> &'bufs mut [&'data [u8]] {
-        assert!(bufs.len() >= 1);
+        assert!(!bufs.is_empty());
 
         let buf = self.buf.as_ref();
         let buf_len = buf.len();
@@ -628,9 +757,9 @@ impl<T: AsRef<[u8]> + AsMut<[u8]>> RefRead for BaseRingBuffer<T> {
         }
     }
 
-    fn get_mut_vectored<'data, 'bufs>(
+    fn get_mut_vectored<'data, 'bufs, const N: usize>(
         &'data mut self,
-        bufs: &'bufs mut MaybeUninit<[&'data mut [u8]; VECTORED_MAX]>,
+        bufs: &'bufs mut MaybeUninit<[&'data mut [u8]; N]>,
     ) -> &'bufs mut [&'data mut [u8]] {
         let bufs = unsafe { bufs.assume_init_mut() };
 
@@ -975,4 +1104,48 @@ mod tests {
         let size = r.read(&mut buf).unwrap();
         assert_eq!(&buf[..size], b"567890ab");
     }
+
+    #[test]
+    fn test_limitbufs() {
+        let mut buf1 = [b'1', b'2', b'3', b'4'];
+        let mut buf2 = [b'5', b'6', b'7', b'8'];
+        let mut buf3 = [b'9', b'0', b'a', b'b'];
+
+        let mut bufs = [buf1.as_slice(), buf2.as_slice(), buf3.as_slice()];
+        {
+            let limited = bufs.limit(7);
+            let limited = limited.as_slice();
+            assert_eq!(limited.len(), 2);
+            assert_eq!(&limited[0], b"1234");
+            assert_eq!(&limited[1], b"567");
+        }
+        assert_eq!(bufs.len(), 3);
+        assert_eq!(&bufs[0], b"1234");
+        assert_eq!(&bufs[1], b"5678");
+        assert_eq!(&bufs[2], b"90ab");
+
+        let mut bufs = [
+            buf1.as_mut_slice(),
+            buf2.as_mut_slice(),
+            buf3.as_mut_slice(),
+        ];
+        {
+            let mut limited = bufs.limit(7);
+            let limited = limited.as_slice();
+            assert_eq!(limited.len(), 2);
+            assert_eq!(&limited[0], b"1234");
+            assert_eq!(&limited[1], b"567");
+        }
+        {
+            let mut limited = bufs.skip(7);
+            let limited = limited.as_slice();
+            assert_eq!(limited.len(), 2);
+            assert_eq!(&limited[0], b"8");
+            assert_eq!(&limited[1], b"90ab");
+        }
+        assert_eq!(bufs.len(), 3);
+        assert_eq!(&bufs[0], b"1234");
+        assert_eq!(&bufs[1], b"5678");
+        assert_eq!(&bufs[2], b"90ab");
+    }
 }
