@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2020 Fanout, Inc.
+ * Copyright (C) 2020-2021 Fanout, Inc.
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -27,61 +27,9 @@ const WHEEL_MAX: usize = WHEEL_LEN - 1;
 const WHEEL_MASK: u64 = (WHEEL_LEN as u64) - 1;
 const TIMEOUT_MAX: u64 = (1 << (WHEEL_BITS * WHEEL_NUM)) - 1;
 
-// count leading zeros
-fn clz64(x: u64) -> u8 {
-    if x == 0 {
-        return 64;
-    }
-
-    let mut t = 1 << 63;
-    let mut r = 0;
-
-    while x & t == 0 {
-        t >>= 1;
-        r += 1;
-    }
-
-    r
-}
-
-// count trailing zeros
-fn ctz64(x: u64) -> u8 {
-    if x == 0 {
-        return 64;
-    }
-
-    let mut t = 1;
-    let mut r = 0;
-
-    while x & t == 0 {
-        t <<= 1;
-        r += 1;
-    }
-
-    r
-}
-
 // find last set
-fn fls64(x: u64) -> u8 {
-    64 - clz64(x)
-}
-
-// rotate left
-fn rotl64(x: u64, count: u8) -> u64 {
-    if count & 0b111111 == 0 {
-        return x;
-    }
-
-    (x << count) | (x >> (64 - count))
-}
-
-// rotate right
-fn rotr64(x: u64, count: u8) -> u64 {
-    if count & 0b111111 == 0 {
-        return x;
-    }
-
-    (x >> count) | (x << (64 - count))
+fn fls64(x: u64) -> u32 {
+    64 - x.leading_zeros()
 }
 
 fn need_resched(curtime: u64, newtime: u64) -> [u64; WHEEL_NUM] {
@@ -94,33 +42,33 @@ fn need_resched(curtime: u64, newtime: u64) -> [u64; WHEEL_NUM] {
 
     let mut elapsed = newtime - curtime;
 
-    for wheel in 0..WHEEL_NUM {
+    for (wheel, item) in result.iter_mut().enumerate() {
         // we only care about the highest bits
         let trunc_bits = (wheel * WHEEL_BITS) as u64;
 
-        let pending;
-
-        if (elapsed >> trunc_bits) > (WHEEL_MAX as u64) {
+        let pending = if (elapsed >> trunc_bits) > (WHEEL_MAX as u64) {
             // all slots need processing
-            pending = !0;
+            !0
         } else {
             let old_slot = (curtime >> trunc_bits) & WHEEL_MASK;
             let new_slot = (newtime >> trunc_bits) & WHEEL_MASK;
 
-            let d = if new_slot >= old_slot {
+            let d = if new_slot > old_slot {
                 new_slot - old_slot
             } else {
                 (WHEEL_LEN as u64) - old_slot + new_slot
             };
 
-            pending = if wheel > 0 {
-                rotl64((1 << d) - 1, old_slot as u8)
+            if d >= WHEEL_LEN as u64 {
+                !0
+            } else if wheel > 0 {
+                ((1 << d) - 1u64).rotate_left(old_slot as u32)
             } else {
-                rotl64((1 << d) - 1, (old_slot + 1) as u8)
-            };
-        }
+                ((1 << d) - 1u64).rotate_left((old_slot + 1) as u32)
+            }
+        };
 
-        result[wheel] = pending;
+        *item = pending;
 
         let finished_bit = if wheel > 0 {
             // higher wheels have completed a full rotation when slot 63 is processed
@@ -143,6 +91,35 @@ fn need_resched(curtime: u64, newtime: u64) -> [u64; WHEEL_NUM] {
     result
 }
 
+#[cfg(test)]
+fn need_resched_simple(curtime: u64, newtime: u64) -> [u64; WHEEL_NUM] {
+    let mut result = [0; WHEEL_NUM];
+
+    // no time elapsed
+    if newtime <= curtime {
+        return result;
+    }
+
+    for curtime in curtime..newtime {
+        for wheel in 0..WHEEL_NUM {
+            let trunc_bits = (wheel * WHEEL_BITS) as u64;
+
+            let old_slot = (curtime >> trunc_bits) & WHEEL_MASK;
+            let new_slot = ((curtime + 1) >> trunc_bits) & WHEEL_MASK;
+
+            if old_slot != new_slot {
+                if wheel > 0 {
+                    result[wheel] |= 1 << old_slot;
+                } else {
+                    result[wheel] |= 1 << new_slot;
+                }
+            }
+        }
+    }
+
+    result
+}
+
 enum InList {
     Wheel(usize, usize),
     Expired,
@@ -173,6 +150,7 @@ impl TimerWheel {
         }
     }
 
+    #[allow(clippy::result_unit_err)]
     pub fn add(&mut self, expires: u64, user_data: usize) -> Result<usize, ()> {
         if self.nodes.len() == self.nodes.capacity() {
             return Err(());
@@ -215,7 +193,7 @@ impl TimerWheel {
         self.nodes.remove(key);
     }
 
-    pub fn timeout(&mut self) -> Option<u64> {
+    pub fn timeout(&self) -> Option<u64> {
         if !self.expired.is_empty() {
             return Some(0);
         }
@@ -230,13 +208,13 @@ impl TimerWheel {
             if self.pending[wheel] != 0 {
                 let slot = ((self.curtime >> trunc_bits) & WHEEL_MASK) as usize;
 
-                let pending = rotr64(self.pending[wheel], slot as u8);
+                let pending = self.pending[wheel].rotate_right(slot as u32);
 
                 // for higher order wheels, timeouts are one step in the future
                 let offset = if wheel > 0 { 1 } else { 0 };
 
                 // pending is guaranteed to be non-zero
-                let t = ((ctz64(pending) as u64) + offset) << trunc_bits;
+                let t = ((pending.trailing_zeros() as u64) + offset) << trunc_bits;
 
                 // reduce by how much lower wheels have progressed
                 let t = t - (relmask & self.curtime);
@@ -264,13 +242,11 @@ impl TimerWheel {
 
         let mut l = list::List::default();
 
-        for wheel in 0..WHEEL_NUM {
-            let pending = need[wheel];
-
+        for (wheel, &pending) in need.iter().enumerate() {
             // loop as long as we still have slots to process
             while pending & self.pending[wheel] != 0 {
                 // get rightmost (earliest) slot that needs processing
-                let slot = ctz64(pending & self.pending[wheel]) as usize;
+                let slot = (pending & self.pending[wheel]).trailing_zeros() as usize;
 
                 // move the timers out
                 l.concat(&mut self.nodes, &mut self.wheel[wheel][slot]);
@@ -393,27 +369,17 @@ mod tests {
         result
     }
 
-    // reverse order of array for human readability
-    fn rev(a: [u64; WHEEL_NUM]) -> [u64; WHEEL_NUM] {
-        [a[3], a[2], a[1], a[0]]
-    }
+    // convert wheel ranges of the form "x:x:x:x", where each part is a range
+    fn r2w(s: &str) -> [u64; WHEEL_NUM] {
+        let mut result = [0; WHEEL_NUM];
 
-    #[test]
-    fn test_clz() {
-        assert_eq!(clz64(0), 64);
-        assert_eq!(clz64(0b1), 63);
-        assert_eq!(clz64(0b10), 62);
-        assert_eq!(clz64(0x4000000000000000), 1);
-        assert_eq!(clz64(0x8000000000000000), 0);
-    }
+        for (i, part) in s.rsplit(":").enumerate() {
+            if !part.is_empty() {
+                result[i] = r2b(part);
+            }
+        }
 
-    #[test]
-    fn test_ctz() {
-        assert_eq!(ctz64(0), 64);
-        assert_eq!(ctz64(0b1), 0);
-        assert_eq!(ctz64(0b10), 1);
-        assert_eq!(ctz64(0x4000000000000000), 62);
-        assert_eq!(ctz64(0x8000000000000000), 63);
+        result
     }
 
     #[test]
@@ -425,24 +391,6 @@ mod tests {
         assert_eq!(fls64(0x8000000000000000), 64);
     }
 
-    #[test]
-    fn test_rotl() {
-        assert_eq!(rotl64(0x0ffff00000000000, 0), 0x0ffff00000000000);
-        assert_eq!(rotl64(0x0ffff00000000000, 64), 0x0ffff00000000000);
-        assert_eq!(rotl64(0x0ffff00000000000, 4), 0xffff000000000000);
-        assert_eq!(rotl64(0x0ffff00000000000, 8), 0xfff000000000000f);
-        assert_eq!(rotl64(0x0ffff00000000000, 16), 0xf000000000000fff);
-    }
-
-    #[test]
-    fn test_rotr() {
-        assert_eq!(rotr64(0x00000000000ffff0, 0), 0x00000000000ffff0);
-        assert_eq!(rotr64(0x00000000000ffff0, 64), 0x00000000000ffff0);
-        assert_eq!(rotr64(0x00000000000ffff0, 4), 0x000000000000ffff);
-        assert_eq!(rotr64(0x00000000000ffff0, 8), 0xf000000000000fff);
-        assert_eq!(rotr64(0x00000000000ffff0, 16), 0xfff000000000000f);
-    }
-
     #[test]
     fn test_sched() {
         let mut w = TimerWheel::new(10);
@@ -477,62 +425,66 @@ mod tests {
 
     #[test]
     fn test_need_resched() {
-        // 00:00 -> 00:00 []:[]
-        assert_eq!(need_resched(ts("00:00"), ts("00:00")), [0, 0, 0, 0]);
+        struct Test {
+            curtime: &'static str,
+            newtime: &'static str,
+            expected: &'static str,
+        }
 
-        // 00:00 -> 00:01 []:[01-01]
-        assert_eq!(
-            need_resched(ts("00:00"), ts("00:01")),
-            rev([0, 0, 0, r2b("01-01")])
-        );
+        fn t(curtime: &'static str, newtime: &'static str, expected: &'static str) -> Test {
+            Test {
+                curtime,
+                newtime,
+                expected,
+            }
+        }
 
-        // 00:01 -> 00:02 []:[02-02]
-        assert_eq!(
-            need_resched(ts("00:01"), ts("00:02")),
-            rev([0, 0, 0, r2b("02-02")])
-        );
+        let table = [
+            t("00:00", "00:00", ""),
+            t("00:00", "00:01", "01-01"),
+            t("00:01", "00:02", "02-02"),
+            t("00:02", "00:63", "03-63"),
+            t("00:63", "01:00", "00-00:00-00"),
+            t("01:00", "01:02", "01-02"),
+            t("01:02", "05:01", "01-04:00-63"),
+            t("05:01", "05:02", "02-02"),
+            t("05:02", "06:01", "05-05:03-01"),
+            t("00:63:63", "01:00:00", "00-00:63-63:00-00"),
+            t("08:00:00", "08:01:00", "00-00:00-63"),
+            t("04:00:02", "05:01:00", "04-04:00-63:00-63"),
+            t("04:01:02", "05:00:00", "04-04:01-63:00-63"),
+            t("04:00:03", "05:00:00", "04-04:00-63:00-63"),
+            t("04:00:02", "05:00:00", "04-04:00-63:00-63"),
+            t("08:00:19", "08:62:63", "00-61:00-63"),
+            t("08:00:19", "08:63:63", "00-62:00-63"),
+            t("09:00:00", "09:63:62", "00-62:00-63"),
+        ];
 
-        // 00:02 -> 00:63 []:[03-63]
-        assert_eq!(
-            need_resched(ts("00:02"), ts("00:63")),
-            rev([0, 0, 0, r2b("03-63")])
-        );
+        for (row, t) in table.iter().enumerate() {
+            let curtime = ts(t.curtime);
+            let newtime = ts(t.newtime);
+            let expected = r2w(t.expected);
 
-        // 00:63 -> 01:00 [00-00]:[00-00]
-        assert_eq!(
-            need_resched(ts("00:63"), ts("01:00")),
-            rev([0, 0, r2b("00-00"), r2b("00-00")])
-        );
+            // ensure the simple algorithm returns what we expect
+            assert_eq!(
+                need_resched_simple(curtime, newtime),
+                expected,
+                "row={} curtime={} newtime={}",
+                row,
+                curtime,
+                newtime
+            );
 
-        // 01:00 -> 01:02 []:[01-02]
-        assert_eq!(
-            need_resched(ts("01:00"), ts("01:02")),
-            rev([0, 0, 0, r2b("01-02")])
-        );
-
-        // 01:02 -> 05:01 [01-04]:[00-63]
-        assert_eq!(
-            need_resched(ts("01:02"), ts("05:01")),
-            rev([0, 0, r2b("01-04"), r2b("00-63")])
-        );
-
-        // 05:01 -> 05:02 []:[02-02]
-        assert_eq!(
-            need_resched(ts("05:01"), ts("05:02")),
-            rev([0, 0, 0, r2b("02-02")])
-        );
-
-        // 05:02 -> 06:01 [05-05]:[03-01]
-        assert_eq!(
-            need_resched(ts("05:02"), ts("06:01")),
-            rev([0, 0, r2b("05-05"), r2b("03-01")])
-        );
-
-        // 00:63:63 -> 01:00:00 [00-00]:[63-63]:[00-00]
-        assert_eq!(
-            need_resched(ts("00:63:63"), ts("01:00:00")),
-            rev([0, r2b("00-00"), r2b("63-63"), r2b("00-00")])
-        );
+            // ensure the optimized algorithm returns matching results
+            assert_eq!(
+                need_resched(curtime, newtime),
+                expected,
+                "row={} curtime={} newtime={}",
+                row,
+                curtime,
+                newtime
+            );
+        }
     }
 
     #[test]
