@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2020-2022 Fanout, Inc.
+ * Copyright (C) 2020-2023 Fanout, Inc.
  *
  * Licensed under the Apache License, Version 2.0 (the "License");
  * you may not use this file except in compliance with the License.
@@ -71,9 +71,8 @@ impl fmt::Display for Bufs<'_> {
         let data: Vec<u8> = self
             .data
             .iter()
-            .map(|v| *v)
-            .flatten()
-            .map(|v| *v)
+            .flat_map(|v| *v)
+            .copied()
             .take(LOG_CONTENT_MAX + 1)
             .collect();
 
@@ -124,6 +123,7 @@ pub fn read_header(buf: &[u8]) -> Result<FrameInfo, io::Error> {
 
     let b1 = buf[1] & 0x7f;
 
+    #[allow(clippy::comparison_chain)]
     let (mut hsize, psize) = if b1 < (PSIZE_3BYTE as u8) {
         (2, b1 as usize)
     } else if b1 == (PSIZE_3BYTE as u8) {
@@ -217,12 +217,20 @@ pub fn write_header(
     }
 }
 
-fn apply_mask(buf: &mut [u8], mask: [u8; 4], offset: usize) {
+pub fn apply_mask(buf: &mut [u8], mask: [u8; 4], offset: usize) {
     for (i, c) in buf.iter_mut().enumerate() {
         *c ^= mask[(offset + i) % 4];
     }
 }
 
+pub fn apply_mask_vectored(bufs: &mut [&mut [u8]], mask: [u8; 4], offset: usize) {
+    let mut count = 0;
+    for buf in bufs {
+        apply_mask(buf, mask, offset + count);
+        count += buf.len();
+    }
+}
+
 fn parse_empty(s: &str, dest: &mut bool) -> Result<(), io::Error> {
     // must not be set yet and value must be empty
     if *dest || !s.is_empty() {
@@ -255,7 +263,7 @@ fn parse_bits(s: &str, dest: &mut Option<u8>, default: Option<u8>) -> Result<(),
     };
 
     // number must be between 8 and 15, inclusive
-    if x >= 8 && x <= 15 {
+    if (8..=15).contains(&x) {
         *dest = Some(x);
         return Ok(());
     }
@@ -297,6 +305,7 @@ impl PerMessageDeflateConfig {
         })
     }
 
+    #[allow(clippy::result_unit_err)]
     pub fn create_response(&self) -> Result<Self, ()> {
         // we don't support non-default server_max_window_bits
         if self.server_max_window_bits != DEFAULT_MAX_WINDOW_BITS {
@@ -315,6 +324,16 @@ impl PerMessageDeflateConfig {
         })
     }
 
+    #[allow(clippy::result_unit_err)]
+    pub fn check_response(&self) -> Result<(), ()> {
+        // we don't support non-default client_max_window_bits
+        if self.client_max_window_bits != DEFAULT_MAX_WINDOW_BITS {
+            return Err(());
+        }
+
+        Ok(())
+    }
+
     pub fn serialize<W: Write>(&self, w: &mut W) -> Result<(), io::Error> {
         if self.client_no_context_takeover {
             write!(w, "; client_no_context_takeover")?;
@@ -364,6 +383,7 @@ impl<T: Copy, const N: usize> ArrayVecExt<T> for ArrayVec<T, N> {
     fn resize(&mut self, new_len: usize, value: T) {
         assert!(new_len <= self.capacity());
 
+        #[allow(clippy::comparison_chain)]
         if new_len > self.len() {
             let old_len = self.len();
             unsafe {
@@ -392,9 +412,10 @@ pub struct DeflateEncoder {
     end: bool,
 }
 
+#[allow(clippy::new_without_default)]
 impl DeflateEncoder {
     pub fn new() -> Self {
-        let mut enc = Box::new(deflate::core::CompressorOxide::default());
+        let mut enc = Box::<deflate::core::CompressorOxide>::default();
 
         enc.set_format_and_level(
             DataFormat::Raw,
@@ -517,7 +538,7 @@ impl DeflateEncoder {
                 // keep back the ending bytes in next_buf
                 assert!(next_buf.is_empty());
                 let keep = cmp::min(ENC_NEXT_BUF_SIZE, dest.len());
-                next_buf.write(&dest[(dest.len() - keep)..]).unwrap();
+                next_buf.write_all(&dest[(dest.len() - keep)..]).unwrap();
 
                 let written = dest.len() - keep;
 
@@ -591,7 +612,7 @@ impl DeflateEncoder {
                 }
 
                 // append tmp to next_buf
-                next_buf.write(tmp.as_ref()).unwrap();
+                next_buf.write_all(tmp.as_ref()).unwrap();
 
                 (result.bytes_consumed, written, maybe_more)
             };
@@ -621,6 +642,7 @@ pub struct DeflateDecoder {
     suffix_pos: Option<usize>,
 }
 
+#[allow(clippy::new_without_default)]
 impl DeflateDecoder {
     pub fn new() -> Self {
         Self {
@@ -647,7 +669,7 @@ impl Decoder for DeflateDecoder {
         dest: &mut [u8],
     ) -> Result<(usize, usize, bool), io::Error> {
         let (consumed, mut written) = if self.suffix_pos.is_none() {
-            let result = inflate(&mut self.dec, &src, dest, MZFlush::None);
+            let result = inflate(&mut self.dec, src, dest, MZFlush::None);
 
             match result.status {
                 Ok(MZStatus::Ok) => {}
@@ -805,19 +827,37 @@ where
         assert!(masked >= read);
         masked -= read;
 
-        let mut bufs_arr = MaybeUninit::uninit();
-        let bufs = src.get_mut_vectored(&mut bufs_arr).limit(masked);
+        let mut bufs_arr = MaybeUninit::<[&mut [u8]; VECTORED_MAX]>::uninit();
+        let mut bufs = src.get_mut_vectored(&mut bufs_arr).limit(masked);
 
-        let mut count = 0;
-        for buf in bufs {
-            apply_mask(buf, mask, mask_offset + read + count);
-            count += buf.len();
-        }
+        apply_mask_vectored(bufs.as_slice(), mask, mask_offset + read);
     }
 
     Ok((read, written, output_end))
 }
 
+// mask src, then call f(), then unmask src. f() returns (skip_unmask, R)
+fn with_mask<F, R>(src: &mut [&mut [u8]], mask: Option<[u8; 4]>, mask_offset: usize, f: F) -> R
+where
+    F: FnOnce(&mut [&mut [u8]]) -> (usize, R),
+{
+    if let Some(mask) = mask {
+        apply_mask_vectored(src, mask, mask_offset);
+    }
+
+    let (skip_unmask, ret) = f(src);
+
+    if let Some(mask) = mask {
+        apply_mask_vectored(
+            src.skip(skip_unmask).as_slice(),
+            mask,
+            mask_offset + skip_unmask,
+        );
+    }
+
+    ret
+}
+
 #[cfg(test)]
 pub struct Frame<'a> {
     pub opcode: u8,
@@ -864,8 +904,9 @@ impl From<io::Error> for Error {
 }
 
 struct SendingFrame {
-    header: [u8; HEADER_SIZE_MAX],
-    header_len: usize,
+    opcode: u8,
+    header: ArrayVec<u8, HEADER_SIZE_MAX>,
+    payload_size: usize,
     sent: usize,
 }
 
@@ -873,6 +914,8 @@ struct SendingMessage {
     opcode: u8,
     mask: Option<[u8; 4]>,
     frame_sent: bool,
+    end_len: Option<usize>,
+    enc_output_end: bool,
 }
 
 struct ReceivingMessage {
@@ -882,8 +925,8 @@ struct ReceivingMessage {
 }
 
 struct Sending {
-    frame: Option<SendingFrame>,
-    message: Option<SendingMessage>,
+    frame: RefCell<Option<SendingFrame>>,
+    message: RefCell<Option<SendingMessage>>,
 }
 
 struct Receiving {
@@ -900,29 +943,28 @@ struct DeflateState<T> {
 
 pub struct Protocol<T> {
     state: Cell<State>,
-    sending: RefCell<Sending>,
+    sending: Sending,
     receiving: RefCell<Receiving>,
     deflate_state: Option<RefCell<DeflateState<T>>>,
 }
 
-impl<'buf, T: AsRef<[u8]> + AsMut<[u8]>> Protocol<T> {
-    pub fn new(deflate_config: Option<(PerMessageDeflateConfig, BaseRingBuffer<T>)>) -> Self {
-        let deflate_state = match deflate_config {
-            Some((config, enc_buf)) => Some(RefCell::new(DeflateState {
+impl<T: AsRef<[u8]> + AsMut<[u8]>> Protocol<T> {
+    pub fn new(deflate_config: Option<(bool, BaseRingBuffer<T>)>) -> Self {
+        let deflate_state = deflate_config.map(|(allow_takeover, enc_buf)| {
+            RefCell::new(DeflateState {
                 enc: DeflateEncoder::new(),
                 dec: DeflateDecoder::new(),
-                allow_takeover: !config.server_no_context_takeover,
+                allow_takeover,
                 enc_buf,
-            })),
-            None => None,
-        };
+            })
+        });
 
         Self {
             state: Cell::new(State::Connected),
-            sending: RefCell::new(Sending {
-                frame: None,
-                message: None,
-            }),
+            sending: Sending {
+                frame: RefCell::new(None),
+                message: RefCell::new(None),
+            },
             receiving: RefCell::new(Receiving {
                 frame: None,
                 message: None,
@@ -935,85 +977,117 @@ impl<'buf, T: AsRef<[u8]> + AsMut<[u8]>> Protocol<T> {
         self.state.get()
     }
 
-    pub fn send_frame(
+    pub fn send_frame<W: Write>(
         &self,
-        writer: &mut dyn Write,
+        writer: &mut W,
         opcode: u8,
-        src: &[&[u8]],
+        src: &mut [&mut [u8]],
         fin: bool,
+        rsv1: bool,
         mask: Option<[u8; 4]>,
-        compression_mode: CompressionMode,
     ) -> Result<usize, Error> {
         assert!(self.state.get() == State::Connected || self.state.get() == State::PeerClosed);
 
-        let sending = &mut *self.sending.borrow_mut();
+        let sending_frame = &mut *self.sending.frame.borrow_mut();
 
         let mut src_len = 0;
         for buf in src.iter() {
             src_len += buf.len();
         }
 
-        if sending.frame.is_none() {
-            let rsv1 = compression_mode == CompressionMode::Compressed;
-            let mut h = [0; HEADER_SIZE_MAX];
+        if sending_frame.is_none() {
+            let mut header = ArrayVec::from([0; HEADER_SIZE_MAX]);
 
-            let size = write_header(fin, rsv1, opcode, src_len, mask, &mut h[..])?;
+            let size = write_header(fin, rsv1, opcode, src_len, mask, &mut header)?;
+            header.truncate(size);
 
-            sending.frame = Some(SendingFrame {
-                header: h,
-                header_len: size,
+            *sending_frame = Some(SendingFrame {
+                opcode,
+                header,
+                payload_size: src_len,
                 sent: 0,
             });
         }
 
-        let sending_frame = sending.frame.as_mut().unwrap();
+        let frame = sending_frame.as_mut().unwrap();
+        let header = frame.header.as_slice();
+        let frame_size = header.len() + frame.payload_size;
 
-        let header = &sending_frame.header[..sending_frame.header_len];
+        let (header_remaining, payload_sent) = if frame.sent < header.len() {
+            (header.len() - frame.sent, 0)
+        } else {
+            (0, frame.sent - header.len())
+        };
 
-        let total = header.len() + src_len;
+        assert!(payload_sent <= frame.payload_size);
 
-        let mut out_arr = [&b""[..]; VECTORED_MAX];
-        let mut out_arr_len = 0;
+        let mut src = src.limit(frame.payload_size - payload_sent);
+        let src = src.as_slice();
 
-        out_arr[0] = header;
-        out_arr_len += 1;
+        // to avoid copying, we apply the mask directly to the input
+        // buffer and then revert it on any bytes that weren't written.
+        // in the best case, all bytes will be written with nothing to
+        // revert. in the worst case, nothing will be written and all
+        // the bytes will be reverted
+        let size = with_mask(src, mask, payload_sent, |src| {
+            let mut out = ArrayVec::<&[u8], VECTORED_MAX>::new();
 
-        for buf in src.iter() {
-            out_arr[out_arr_len] = buf;
-            out_arr_len += 1;
-        }
+            if header_remaining > 0 {
+                out.push(&header[frame.sent..]);
+            }
 
-        let out = &out_arr[..out_arr_len];
-        let size = write_vectored_offset(writer, out, sending_frame.sent)?;
+            for buf in src.iter() {
+                out.push(*buf);
+            }
 
-        if log_enabled!(log::Level::Trace) {
-            trace!("OUT sock {}", Bufs::new(out));
-        }
+            let ret = write_vectored_offset(writer, out.as_slice(), 0);
 
-        sending_frame.sent += size;
+            if log_enabled!(log::Level::Trace) {
+                trace!("OUT sock {} -> {:?}", Bufs::new(out.as_slice()), ret);
+            }
 
-        if sending_frame.sent < total {
+            let size = match &ret {
+                Ok(size) => *size,
+                Err(_) => 0,
+            };
+
+            let skip_unmask = size.saturating_sub(header_remaining);
+
+            (skip_unmask, ret)
+        })?;
+
+        frame.sent += size;
+
+        assert!(frame.sent <= frame_size);
+
+        if frame.sent < header.len() {
             return Ok(0);
         }
 
-        sending.frame = None;
+        let payload_written = (frame.sent - header.len()) - payload_sent;
 
-        if opcode == OPCODE_CLOSE {
-            if self.state.get() == State::PeerClosed {
-                self.state.set(State::Finished);
-            } else {
-                self.state.set(State::Closing);
+        if frame.sent == frame_size {
+            let opcode = frame.opcode;
+
+            *sending_frame = None;
+
+            if opcode == OPCODE_CLOSE {
+                if self.state.get() == State::PeerClosed {
+                    self.state.set(State::Finished);
+                } else {
+                    self.state.set(State::Closing);
+                }
             }
         }
 
-        Ok(src_len)
+        Ok(payload_written)
     }
 
     // on success, it's up to the caller to advance the buffer by frame.data.len()
     #[cfg(test)]
-    pub fn recv_frame(
+    pub fn recv_frame<'buf, R: RefRead>(
         &mut self,
-        rbuf: &'buf mut dyn RefRead,
+        rbuf: &'buf mut R,
     ) -> Option<Result<Frame<'buf>, Error>> {
         assert!(self.state.get() == State::Connected || self.state.get() == State::Closing);
 
@@ -1061,40 +1135,62 @@ impl<'buf, T: AsRef<[u8]> + AsMut<[u8]>> Protocol<T> {
     }
 
     pub fn is_sending_message(&self) -> bool {
-        self.sending.borrow().message.is_some()
+        self.sending.message.borrow().is_some()
     }
 
     pub fn send_message_start(&self, opcode: u8, mask: Option<[u8; 4]>) {
         assert!(self.state.get() == State::Connected || self.state.get() == State::PeerClosed);
 
-        let sending = &mut *self.sending.borrow_mut();
+        let sending_message = &mut *self.sending.message.borrow_mut();
 
-        assert_eq!(sending.message.is_some(), false);
+        assert!(sending_message.is_none());
 
-        sending.message = Some(SendingMessage {
+        *sending_message = Some(SendingMessage {
             opcode,
             mask,
             frame_sent: false,
+            end_len: None,
+            enc_output_end: false,
         });
     }
 
-    pub fn send_message_content(
+    // returns (bytes read, done)
+    // note: when compression is used, bytes may be buffered in the encoder
+    // and may not be flushed to the writer until the encoder's buffer is
+    // full or the input ends
+    pub fn send_message_content<W: Write>(
         &self,
-        writer: &mut dyn Write,
-        src: &[&[u8]],
+        writer: &mut W,
+        src: &mut [&mut [u8]],
         end: bool,
     ) -> Result<(usize, bool), Error> {
         assert!(self.state.get() == State::Connected || self.state.get() == State::PeerClosed);
 
-        let sending = self.sending.borrow();
+        let mut sending_message = self.sending.message.borrow_mut();
 
-        let msg = sending.message.as_ref().unwrap();
+        let msg = sending_message.as_mut().unwrap();
 
         let mut src_len = 0;
         for buf in src.iter() {
             src_len += buf.len();
         }
 
+        if let Some(end_len) = msg.end_len {
+            // once the caller has passed end=true, it must continue to pass
+            // end=true in all subsequent calls until this method returns
+            // done
+            assert!(end);
+
+            // once the caller has passed end=true, it must continue to
+            // provide the expected number of src bytes in all subsequent
+            // calls until this method returns done
+            assert_eq!(src_len, end_len);
+        } else if end {
+            // when the caller passes end=true, note the number of src bytes
+            // provided
+            msg.end_len = Some(src_len);
+        }
+
         let is_control = msg.opcode & 0x08 != 0;
 
         // control frames (ping, pong, close) must have a small payload length
@@ -1109,84 +1205,82 @@ impl<'buf, T: AsRef<[u8]> + AsMut<[u8]>> Protocol<T> {
             msg.opcode
         };
 
-        let fin = if let Some(f) = &sending.frame {
-            f.header[0] & 0x80 != 0
-        } else {
-            end
-        };
-
-        let mask = msg.mask;
-
-        drop(sending);
-
-        let read = match &self.deflate_state {
+        let (read, sent_all) = match &self.deflate_state {
             Some(state) if !is_control => {
                 let state = &mut *state.borrow_mut();
 
                 let mut read = 0;
-                let mut output_end = false;
 
-                if src_len > 0 {
-                    for (i, buf) in src.iter().enumerate() {
-                        // only set end on the last buf
-                        let end = end && (i == src.len() - 1);
+                if !msg.enc_output_end {
+                    if src_len > 0 {
+                        for (i, buf) in src.iter().enumerate() {
+                            // only set end on the last buf
+                            let end = end && (i == src.len() - 1);
 
+                            let dest = &mut state.enc_buf;
+
+                            let (r, oe) = state.enc.encode_to_ringbuffer(buf, end, dest)?;
+
+                            read += r;
+                            msg.enc_output_end = oe;
+
+                            if r < buf.len() || oe {
+                                break;
+                            }
+                        }
+                    } else {
                         let dest = &mut state.enc_buf;
 
-                        let (r, oe) = state.enc.encode_to_ringbuffer(buf, end, dest)?;
+                        let (_, oe) = state.enc.encode_to_ringbuffer(&[], end, dest)?;
 
-                        read += r;
-                        output_end = oe;
-
-                        if r < buf.len() || oe {
-                            break;
-                        }
+                        msg.enc_output_end = oe;
                     }
-                } else {
-                    let dest = &mut state.enc_buf;
-
-                    let (r, oe) = state.enc.encode_to_ringbuffer(&[], end, dest)?;
-
-                    read += r;
-                    output_end = oe;
                 }
 
                 // we should never get EOS if there are bytes left to send
-                assert!(!output_end || read == src_len);
+                assert!(!msg.enc_output_end || read == src_len);
 
-                if state.enc_buf.read_avail() > 0 || output_end {
-                    // send_frame adds 1 element to vector
-                    let mut bufs_arr = [&b""[..]; VECTORED_MAX - 1];
-                    let bufs = state.enc_buf.get_ref_vectored(&mut bufs_arr);
-
-                    let size = self.send_frame(
-                        writer,
-                        opcode,
-                        bufs,
-                        output_end,
-                        mask,
-                        CompressionMode::Compressed,
-                    )?;
-
-                    state.enc_buf.read_commit(size);
+                if let Some(end_len) = &mut msg.end_len {
+                    *end_len -= read;
                 }
 
-                read
+                let mut sent_all = false;
+
+                // only attempt to write if we have no consumed byte count
+                // to report, so that if the write returns an error
+                // (including WouldBlock) we can propagate the error without
+                // data loss
+                if read == 0 && (state.enc_buf.read_avail() > 0 || msg.enc_output_end) {
+                    // send_frame adds 1 element to vector
+                    let mut bufs_arr = MaybeUninit::<[&mut [u8]; VECTORED_MAX - 1]>::uninit();
+                    let bufs = state.enc_buf.get_mut_vectored(&mut bufs_arr);
+
+                    // set on first frame
+                    let rsv1 = opcode != OPCODE_CONTINUATION;
+
+                    let size =
+                        self.send_frame(writer, opcode, bufs, msg.enc_output_end, rsv1, msg.mask)?;
+
+                    state.enc_buf.read_commit(size);
+
+                    msg.frame_sent = true;
+
+                    sent_all = msg.enc_output_end && state.enc_buf.read_avail() == 0;
+                }
+
+                (read, sent_all)
+            }
+            _ => {
+                let read = self.send_frame(writer, opcode, src, end, false, msg.mask)?;
+
+                msg.frame_sent = true;
+
+                (read, end && read == src_len)
             }
-            _ => self.send_frame(
-                writer,
-                opcode,
-                src,
-                fin,
-                mask,
-                CompressionMode::Uncompressed,
-            )?,
         };
 
-        let mut sending = self.sending.borrow_mut();
-
-        if sending.frame.is_none() && fin {
-            sending.message = None;
+        if sent_all && self.sending.frame.borrow().is_none() {
+            *sending_message = None;
 
             if let Some(state) = &self.deflate_state {
                 let mut state = state.borrow_mut();
@@ -1195,19 +1289,16 @@ impl<'buf, T: AsRef<[u8]> + AsMut<[u8]>> Protocol<T> {
                     state.enc.reset();
                 }
             }
-        } else {
-            let msg = sending.message.as_mut().unwrap();
-            msg.frame_sent = true;
         }
 
-        let done = sending.message.is_none();
+        let done = sending_message.is_none();
 
         Ok((read, done))
     }
 
-    pub fn recv_message_content(
+    pub fn recv_message_content<R: RefRead>(
         &self,
-        rbuf: &mut dyn RefRead,
+        rbuf: &mut R,
         dest: &mut [u8],
     ) -> Option<Result<(u8, usize, bool), Error>> {
         assert!(self.state.get() == State::Connected || self.state.get() == State::Closing);
@@ -1264,7 +1355,7 @@ impl<'buf, T: AsRef<[u8]> + AsMut<[u8]>> Protocol<T> {
         let fi = receiving.frame.as_ref().unwrap();
         let msg = receiving.message.as_mut().unwrap();
 
-        let (written, output_end) = if msg.compression_mode == CompressionMode::Compressed {
+        let (written, frame_read_end) = if msg.compression_mode == CompressionMode::Compressed {
             let state = match &self.deflate_state {
                 Some(state) => state,
                 None => return Some(Err(Error::CompressionError)),
@@ -1275,7 +1366,7 @@ impl<'buf, T: AsRef<[u8]> + AsMut<[u8]>> Protocol<T> {
             let left = fi.payload_size - msg.frame_payload_read;
             let limit = cmp::min(left, rbuf.len());
 
-            let end = (msg.frame_payload_read + limit) == fi.payload_size;
+            let end = ((msg.frame_payload_read + limit) == fi.payload_size) && fi.fin;
 
             let (read, written, output_end) = match unmask_and_decode(
                 rbuf,
@@ -1292,12 +1383,20 @@ impl<'buf, T: AsRef<[u8]> + AsMut<[u8]>> Protocol<T> {
 
             msg.frame_payload_read += read;
 
-            // no output progress means we need more input
-            if written == 0 && !output_end {
+            let frame_read_end = if fi.fin {
+                // finish final frame only when we hit EOS
+                (msg.frame_payload_read == fi.payload_size) && output_end
+            } else {
+                msg.frame_payload_read == fi.payload_size
+            };
+
+            if !frame_read_end && written == 0 && rbuf.len() == 0 {
+                // if there's no progress to report and nothing left to read
+                // then we need more input
                 return None;
             }
 
-            (written, output_end)
+            (written, frame_read_end)
         } else {
             let buf = rbuf.get_ref();
 
@@ -1308,7 +1407,7 @@ impl<'buf, T: AsRef<[u8]> + AsMut<[u8]>> Protocol<T> {
 
             let left = fi.payload_size - msg.frame_payload_read;
 
-            if left > 0 && buf.len() == 0 {
+            if left > 0 && buf.is_empty() {
                 return None;
             }
 
@@ -1336,7 +1435,7 @@ impl<'buf, T: AsRef<[u8]> + AsMut<[u8]>> Protocol<T> {
         let opcode = msg.opcode;
         let fin = fi.fin;
 
-        if output_end {
+        if frame_read_end {
             receiving.frame = None;
 
             if fin {
@@ -1368,7 +1467,7 @@ pub mod testutil {
 
     pub struct BenchSendMessage {
         use_deflate: bool,
-        content: Vec<u8>,
+        content: RefCell<Vec<u8>>,
     }
 
     impl BenchSendMessage {
@@ -1380,7 +1479,7 @@ pub mod testutil {
 
             Self {
                 use_deflate,
-                content,
+                content: RefCell::new(content),
             }
         }
 
@@ -1388,10 +1487,7 @@ pub mod testutil {
             let deflate_config = if self.use_deflate {
                 let tmp = Rc::new(TmpBuffer::new(256));
 
-                Some((
-                    PerMessageDeflateConfig::default(),
-                    RingBuffer::new(256, &tmp),
-                ))
+                Some((true, RingBuffer::new(256, &tmp)))
             } else {
                 None
             };
@@ -1404,7 +1500,7 @@ pub mod testutil {
 
         pub fn run(&self, args: &mut BenchSendMessageArgs) {
             let p = &mut args.protocol;
-            let src = &self.content;
+            let src = &mut *self.content.borrow_mut();
             let dest = &mut args.dest;
 
             p.send_message_start(OPCODE_TEXT, None);
@@ -1413,7 +1509,7 @@ pub mod testutil {
 
             loop {
                 let (size, done) = p
-                    .send_message_content(dest, &[&src[src_pos..]], true)
+                    .send_message_content(dest, &mut [&mut src[src_pos..]], true)
                     .unwrap();
 
                 src_pos += size;
@@ -1449,10 +1545,7 @@ pub mod testutil {
             let tmp = Rc::new(TmpBuffer::new(16_384));
 
             let deflate_config = if use_deflate {
-                Some((
-                    PerMessageDeflateConfig::default(),
-                    RingBuffer::new(16_384, &tmp),
-                ))
+                Some((true, RingBuffer::new(16_384, &tmp)))
             } else {
                 None
             };
@@ -1461,9 +1554,20 @@ pub mod testutil {
 
             let mut msg = Vec::new();
             p.send_message_start(OPCODE_TEXT, None);
-            let (size, done) = p.send_message_content(&mut msg, &[&content], true).unwrap();
-            assert_eq!(size, content.len());
-            assert_eq!(done, true);
+
+            let mut src_pos = 0;
+
+            loop {
+                let (size, done) = p
+                    .send_message_content(&mut msg, &mut [&mut content[src_pos..]], true)
+                    .unwrap();
+
+                src_pos += size;
+
+                if done {
+                    break;
+                }
+            }
 
             Self {
                 use_deflate,
@@ -1474,10 +1578,7 @@ pub mod testutil {
 
         pub fn init(&self) -> BenchRecvMessageArgs {
             let deflate_config = if self.use_deflate {
-                Some((
-                    PerMessageDeflateConfig::default(),
-                    RingBuffer::new(256, &self.tmp),
-                ))
+                Some((true, RingBuffer::new(256, &self.tmp)))
             } else {
                 None
             };
@@ -1487,8 +1588,7 @@ pub mod testutil {
             let size = rbuf.write(&self.msg).unwrap();
             assert_eq!(size, self.msg.len());
 
-            let mut dest = Vec::with_capacity(16_384);
-            dest.resize(dest.capacity(), 0);
+            let dest = vec![0; 16_384];
 
             BenchRecvMessageArgs {
                 protocol: Protocol::new(deflate_config),
@@ -1527,21 +1627,34 @@ mod tests {
     use super::testutil::*;
     use super::*;
     use crate::buffer::{RingBuffer, TmpBuffer};
+    use std::collections::VecDeque;
     use std::rc::Rc;
 
     struct MyWriter {
         data: Vec<u8>,
+        allow: usize,
     }
 
     impl MyWriter {
         fn new() -> Self {
-            Self { data: Vec::new() }
+            Self {
+                data: Vec::new(),
+                allow: 1024,
+            }
         }
     }
 
     impl Write for MyWriter {
         fn write(&mut self, buf: &[u8]) -> Result<usize, io::Error> {
-            self.data.extend_from_slice(buf.as_ref());
+            if !buf.is_empty() && self.allow == 0 {
+                return Err(io::Error::from(io::ErrorKind::WouldBlock));
+            }
+
+            let size = cmp::min(buf.len(), self.allow);
+            let buf = &buf[..size];
+
+            self.data.extend_from_slice(buf);
+            self.allow -= size;
 
             Ok(buf.len())
         }
@@ -1550,8 +1663,25 @@ mod tests {
             let mut total = 0;
 
             for buf in bufs {
+                if buf.is_empty() {
+                    continue;
+                }
+
+                if self.allow == 0 {
+                    if total == 0 {
+                        return Err(io::Error::from(io::ErrorKind::WouldBlock));
+                    }
+
+                    break;
+                }
+
+                let size = cmp::min(buf.len(), self.allow);
+                let buf = &buf[..size];
+
+                self.data.extend_from_slice(buf);
+                self.allow -= size;
+
                 total += buf.len();
-                self.data.extend_from_slice(buf.as_ref());
             }
 
             Ok(total)
@@ -1562,6 +1692,10 @@ mod tests {
         }
     }
 
+    fn make_buf<const N: usize>(s: &[u8; N]) -> ArrayVec<u8, N> {
+        ArrayVec::from(*s)
+    }
+
     #[test]
     fn test_header_size() {
         assert_eq!(header_size(0, false), 2);
@@ -1725,10 +1859,10 @@ mod tests {
             .send_frame(
                 &mut writer,
                 OPCODE_TEXT,
-                &[b"hello"],
+                &mut [&mut make_buf(b"hello")],
                 true,
+                false,
                 None,
-                CompressionMode::Uncompressed,
             )
             .unwrap();
 
@@ -1737,6 +1871,118 @@ mod tests {
         assert_eq!(p.state(), State::Connected);
     }
 
+    #[test]
+    fn test_send_frame_masked() {
+        let p = Protocol::<[u8; 0]>::new(None);
+
+        let mut buf = make_buf(b"hello");
+
+        let mut writer = MyWriter::new();
+        writer.allow = 0;
+
+        let e = p
+            .send_frame(
+                &mut writer,
+                OPCODE_TEXT,
+                &mut [&mut buf],
+                true,
+                false,
+                Some([0x01, 0x02, 0x03, 0x04]),
+            )
+            .unwrap_err();
+
+        let e = match e {
+            Error::Io(e) => e,
+            _ => panic!("unexpected error type"),
+        };
+
+        assert_eq!(e.kind(), io::ErrorKind::WouldBlock);
+        assert_eq!(&buf, b"hello".as_slice());
+        assert_eq!(p.sending.frame.borrow().is_some(), true);
+
+        writer.allow = 3;
+
+        let size = p
+            .send_frame(
+                &mut writer,
+                OPCODE_TEXT,
+                &mut [&mut buf],
+                true,
+                false,
+                Some([0x01, 0x02, 0x03, 0x04]),
+            )
+            .unwrap();
+
+        let expected = [0x81, 0x85, 0x01];
+
+        assert_eq!(size, 0);
+        assert_eq!(writer.data, expected);
+        assert_eq!(&buf, b"hello".as_slice());
+        assert_eq!(p.sending.frame.borrow().is_some(), true);
+
+        writer.allow = 4;
+
+        let size = p
+            .send_frame(
+                &mut writer,
+                OPCODE_TEXT,
+                &mut [&mut buf],
+                true,
+                false,
+                Some([0x01, 0x02, 0x03, 0x04]),
+            )
+            .unwrap();
+
+        let expected = [0x81, 0x85, 0x01, 0x02, 0x03, 0x04, b'h' ^ 0x01];
+
+        assert_eq!(size, 1);
+        assert_eq!(writer.data, expected);
+        assert_eq!(&buf, [b'h' ^ 0x01, b'e', b'l', b'l', b'o'].as_slice());
+        assert_eq!(p.sending.frame.borrow().is_some(), true);
+
+        writer.allow = 1024;
+
+        let size = p
+            .send_frame(
+                &mut writer,
+                OPCODE_TEXT,
+                &mut [&mut buf[1..]],
+                true,
+                false,
+                Some([0x01, 0x02, 0x03, 0x04]),
+            )
+            .unwrap();
+
+        let expected = [
+            0x81,
+            0x85,
+            0x01,
+            0x02,
+            0x03,
+            0x04,
+            b'h' ^ 0x01,
+            b'e' ^ 0x02,
+            b'l' ^ 0x03,
+            b'l' ^ 0x04,
+            b'o' ^ 0x01,
+        ];
+
+        assert_eq!(size, 4);
+        assert_eq!(writer.data, expected);
+        assert_eq!(
+            &buf,
+            [
+                b'h' ^ 0x01,
+                b'e' ^ 0x02,
+                b'l' ^ 0x03,
+                b'l' ^ 0x04,
+                b'o' ^ 0x01
+            ]
+            .as_slice()
+        );
+        assert_eq!(p.sending.frame.borrow().is_some(), false);
+    }
+
     #[test]
     fn test_send_message() {
         let p = Protocol::<[u8; 0]>::new(None);
@@ -1748,7 +1994,11 @@ mod tests {
         p.send_message_start(OPCODE_TEXT, None);
 
         let (size, done) = p
-            .send_message_content(&mut writer, &[b"hel", b"lo"], true)
+            .send_message_content(
+                &mut writer,
+                &mut [&mut make_buf(b"hel"), &mut make_buf(b"lo")],
+                true,
+            )
             .unwrap();
         assert_eq!(size, 5);
         assert_eq!(done, true);
@@ -1759,7 +2009,7 @@ mod tests {
         p.send_message_start(OPCODE_TEXT, None);
 
         let (size, done) = p
-            .send_message_content(&mut writer, &[b"hello"], false)
+            .send_message_content(&mut writer, &mut [&mut make_buf(b"hello")], false)
             .unwrap();
         assert_eq!(size, 5);
         assert_eq!(done, false);
@@ -1768,7 +2018,7 @@ mod tests {
 
         writer.data.clear();
 
-        let (size, done) = p.send_message_content(&mut writer, &[b""], true).unwrap();
+        let (size, done) = p.send_message_content(&mut writer, &mut [], true).unwrap();
         assert_eq!(size, 0);
         assert_eq!(done, true);
         assert_eq!(writer.data, b"\x80\x00");
@@ -1778,7 +2028,7 @@ mod tests {
         p.send_message_start(OPCODE_PING, None);
 
         let (size, done) = p
-            .send_message_content(&mut writer, &[b"hello"], true)
+            .send_message_content(&mut writer, &mut [&mut make_buf(b"hello")], true)
             .unwrap();
         assert_eq!(size, 5);
         assert_eq!(done, true);
@@ -1788,7 +2038,7 @@ mod tests {
         writer.data.clear();
         p.send_message_start(OPCODE_PING, None);
 
-        let r = p.send_message_content(&mut writer, &[b"hello"], false);
+        let r = p.send_message_content(&mut writer, &mut [&mut make_buf(b"hello")], false);
         assert!(r.is_err());
     }
 
@@ -1925,19 +2175,25 @@ mod tests {
     fn test_send_recv_compressed() {
         let tmp = Rc::new(TmpBuffer::new(1024));
 
-        let p = Protocol::new(Some((
-            PerMessageDeflateConfig::default(),
-            RingBuffer::new(1024, &tmp),
-        )));
+        let p = Protocol::new(Some((true, RingBuffer::new(1024, &tmp))));
 
         let mut writer = MyWriter::new();
 
         p.send_message_start(OPCODE_TEXT, None);
 
         let (size, done) = p
-            .send_message_content(&mut writer, &[b"Hel", b"lo"], true)
+            .send_message_content(
+                &mut writer,
+                &mut [&mut make_buf(b"Hel"), &mut make_buf(b"lo")],
+                true,
+            )
             .unwrap();
         assert_eq!(size, 5);
+        assert_eq!(done, false);
+        assert_eq!(writer.data.is_empty(), true);
+
+        let (size, done) = p.send_message_content(&mut writer, &mut [], true).unwrap();
+        assert_eq!(size, 0);
         assert_eq!(done, true);
         assert_eq!(
             writer.data,
@@ -1946,10 +2202,7 @@ mod tests {
 
         let mut rbuf = io::Cursor::new(writer.data.as_mut());
 
-        let p = Protocol::new(Some((
-            PerMessageDeflateConfig::default(),
-            RingBuffer::new(1024, &tmp),
-        )));
+        let p = Protocol::new(Some((true, RingBuffer::new(1024, &tmp))));
 
         let mut dest = [0; 1024];
 
@@ -1964,6 +2217,86 @@ mod tests {
         assert_eq!(end, true);
     }
 
+    #[test]
+    fn test_send_recv_compressed_fragmented() {
+        let tmp = Rc::new(TmpBuffer::new(1024));
+
+        let p = Protocol::new(Some((true, RingBuffer::new(1024, &tmp))));
+
+        let mut writer = MyWriter::new();
+
+        p.send_message_start(OPCODE_TEXT, None);
+
+        let (size, done) = p
+            .send_message_content(&mut writer, &mut [&mut make_buf(b"hello")], false)
+            .unwrap();
+        assert_eq!(size, 5);
+        assert_eq!(done, false);
+
+        // flush the encoded data
+        {
+            let state = &mut *p.deflate_state.as_ref().unwrap().borrow_mut();
+
+            let (_, output_end) = state
+                .enc
+                .encode_to_ringbuffer(&[], true, &mut state.enc_buf)
+                .unwrap();
+            assert_eq!(output_end, true);
+
+            state.enc_buf.write(&DEFLATE_SUFFIX).unwrap();
+        }
+
+        // send flushed data as first frame
+        let (size, done) = p.send_message_content(&mut writer, &mut [], false).unwrap();
+        assert_eq!(size, 0);
+        assert_eq!(done, false);
+        assert_eq!(writer.data.is_empty(), false);
+
+        // send second frame
+        let (size, done) = p
+            .send_message_content(&mut writer, &mut [&mut make_buf(b" world")], true)
+            .unwrap();
+        assert_eq!(size, 6);
+        assert_eq!(done, false);
+
+        let (size, done) = p.send_message_content(&mut writer, &mut [], true).unwrap();
+        assert_eq!(size, 0);
+        assert_eq!(done, true);
+
+        let p = Protocol::new(Some((true, RingBuffer::new(1024, &tmp))));
+
+        let mut writer_data = VecDeque::from(writer.data);
+        let mut input = Vec::new();
+        let mut result: Vec<u8> = Vec::new();
+
+        // feed one byte at a time
+        loop {
+            let mut rbuf = io::Cursor::new(input.as_mut());
+            let mut dest = [0; 1024];
+
+            let ret = p.recv_message_content(&mut rbuf, &mut dest);
+
+            let read = rbuf.position() as usize;
+            input = input.split_off(read);
+
+            if ret.is_none() {
+                input.push(writer_data.pop_front().unwrap());
+                continue;
+            }
+
+            let (opcode, size, end) = ret.unwrap().unwrap();
+            assert_eq!(opcode, OPCODE_TEXT);
+
+            result.extend(&dest[..size]);
+
+            if end {
+                break;
+            }
+        }
+
+        assert_eq!(result, b"hello world");
+    }
+
     struct LimitedDeflateDecoder {
         dec: DeflateDecoder,
         limit: usize,
@@ -2054,9 +2387,21 @@ mod tests {
         t.run(&mut t.init());
     }
 
+    #[test]
+    fn bench_send_message_with_deflate() {
+        let t = BenchSendMessage::new(true);
+        t.run(&mut t.init());
+    }
+
     #[test]
     fn bench_recv_message() {
         let t = BenchRecvMessage::new(false);
         t.run(&mut t.init());
     }
+
+    #[test]
+    fn bench_recv_message_with_deflate() {
+        let t = BenchRecvMessage::new(true);
+        t.run(&mut t.init());
+    }
 }
